/*
** Copyright (C) 2008-2009 SecurixLive   <dev@securixlive.com>
**
** Copyright (C) 2002-2009 Sourcefire, Inc.
** Copyright (C) 1998-2002 Martin Roesch <roesch@sourcefire.com>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License Version 2 as
** published by the Free Software Foundation.  You may not use, modify or
** distribute this program under any other version of the GNU General
** Public License.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/*
** Description:
**   A dedicated snort unified2 log file parser
**
** Author(s):
**   firnsy <firnsy@securixlive.com>
**   SecurixLive.com Team <dev@securixlive.com>
**
** Comments:
**   Foundation is built upon the Snort 2.8.3 codebase (www.snort.org/dl) with
** ideas stolen liberally from:
**     1. the orginal barnyard (A. Baker, M. Roesch)
**
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <syslog.h>
#include <grp.h>
#include <pwd.h>

#include "barnyard2.h"
#include "getopt.h"
#include "debug.h"
#include "parser.h"
#include "plugbase.h"
#include "spooler.h"
#include "map.h"
#include "timersub.h"
#include "util.h"

extern int errno;

/* exported variables *********************************************************/
u_int8_t				runMode = 0;	/* barnyard2 run mode */
PV						pv;             /* program vars */
PacketCount				pc;		        /* packet count information */
int						thiszone;		/* time zone info */
int						datalink;
char					*progname;      /* name of the program (from argv[0]) */
char					**progargs;
char					*username;
char					*groupname;
unsigned long			userid = 0;
unsigned long			groupid = 0;
struct passwd			*pw;
struct group			*gr;

static struct timeval	starttime;
static struct timeval	endtime;

char					*protocol_names[256];
u_long					netmasks[33];			/* precalculated netmask array */
u_int					snaplen = SNAPLEN;


/*
 * you may need to adjust this on the systems which don't have standard
 * paths defined
 */
#ifndef _PATH_VARRUN
char _PATH_VARRUN[STD_BUF];
#endif

/* locally defined functions **************************************************/
static char *ConfigFileSearch();
static void Restart();
static void PrintVersion();
void LoadDataProcessors();
int PluginsExit();
int PluginsRestart();


/* Signal handler declarations ************************************************/
static void SigTermHandler(int signal);
static void SigIntHandler(int signal);
static void SigQuitHandler(int signal);
static void SigHupHandler(int signal);
static void SigUsrHandler(int signal);

#ifdef CATCH_SEGV
static void SigSegvHandler(int signal);
#else
#include <sys/resource.h>
#endif

/*
 *  Check for SIGHUP and invoke restart to
 *  cleanup
 */
int hup_check()
{
    int quiet_flag = pv.quiet_flag;

    if (pv.usr_signal == SIGHUP)
    {
        pv.quiet_flag = 0;
        LogMessage("*** Caught Hup-Signal\n");
        Restart();
        pv.quiet_flag = quiet_flag;
        pv.usr_signal = 0;
        pv.restart_flag = 1;
        return 1;
    }

    if (pv.cant_hup_signal == SIGHUP)
    {
        pv.quiet_flag = 0;
        LogMessage("Reload via Signal HUP does not work if you aren't root or are chroot'ed\n");
        pv.quiet_flag = quiet_flag;
        pv.cant_hup_signal = 0;
    }

    return 0;
}

/*
 *  Check for signal activity 
 */
static int exit_logged = 0;
int sig_check()
{
        int quiet_flag = pv.quiet_flag;
        pv.quiet_flag = 0;

        switch (pv.exit_signal)
        {
#ifndef DEBUG
            case SIGSEGV:
            case SIGFPE:
                /* Try to exit cleanly -- if this fails and we get
                 * another signal, we'll exit immediately.  */
                CleanExit(1);
                return 2;
                break;
#endif
            case SIGTERM:
                if (!exit_logged)
                {
                    LogMessage("*** Caught Term-Signal\n");
                    exit_logged = 1;
                }
                CleanExit(0);
                return 1;
                break;
            case SIGINT:
                if (!exit_logged)
                {
                    LogMessage("*** Caught Int-Signal\n");
                    exit_logged = 1;
                }
                CleanExit(0);
                return 1;
                break;
            case SIGQUIT:
                if (!exit_logged)
                {
                    LogMessage("*** Caught Quit-Signal\n");
                    exit_logged = 1;
                }
                CleanExit(0);
                return 1;
                break;
            case 0:
            default:
                break;
        }
        pv.exit_signal = 0;

        switch (pv.usr_signal)
        {
            case SIGUSR1:
                LogMessage("*** Caught Usr-Signal\n");
                DropStats(0);
                break;
            case SIGHUP:
                pv.quiet_flag = quiet_flag;
                return 1;
                break;
        }
        pv.quiet_flag = quiet_flag;
        pv.usr_signal = 0;

        return 0;
}

/*
 *
 * Function: main(int, char *)
 *
 * Purpose:  Handle program entry and exit, call main prog sections
 *           This can handle both regular (command-line) style
 *           startup, as well as Win32 Service style startup.
 *
 * Arguments: See command line args in README file
 *
 * Returns: 0 => normal exit, 1 => exit on error
 *
 */
int main(int argc, char* argv[]) 
{
	int					ret = 0;

    ret = BarnyardMain(argc, argv);
	CleanExit(ret);

	return 0;
}

/*
 *
 * Function: BarnyardMain(int, char *)
 *
 * Purpose:  The real place that the program handles entry and exit.  Called
 *           called by main().
 *
 * Arguments: See command line args in README file
 *
 * Returns: 0 => normal exit, 1 => exit on error
 *
 */
int BarnyardMain(int argc, char *argv[])
{
	struct timezone		tz;
	sigset_t 			set;
    sigemptyset(&set);
    sigprocmask(SIG_SETMASK, &set, NULL);

    /* Make this prog behave nicely when signals come along.
     * Windows doesn't like all of these signals, and will
     * set errno for some.  Ignore/reset this error so it
     * doesn't interfere with later checks of errno value.
     */
    signal(SIGTERM, SigTermHandler);    if(errno!=0) errno=0;
    signal(SIGINT, SigIntHandler);      if(errno!=0) errno=0;
    signal(SIGQUIT, SigQuitHandler);    if(errno!=0) errno=0;
    signal(SIGHUP, SigHupHandler);      if(errno!=0) errno=0;
    signal(SIGUSR1, SigUsrHandler);    if(errno!=0) errno=0;
   
#ifdef CATCH_SEGV
    signal(SIGSEGV,SigSegvHandler);  if(errno!=0) errno=0;
    signal(SIGFPE, SigSegvHandler);  if(errno!=0) errno=0;
#else
#ifdef NOCOREFILE
    {
        struct rlimit rlim;
        getrlimit(RLIMIT_CORE, &rlim);
        rlim.rlim_max = 0;
        setrlimit(RLIMIT_CORE, &rlim);
    }
#endif
#endif

    /*
     * set a global ptr to the program name so other functions can tell what
     * the program name is
     */
    progname = argv[0];
    progargs = argv;

    memset(&pv, 0, sizeof(PV));

    /* not using a waldo file by default */
    pv.waldo.state = 0;

    /* provide (limited) status messages by default */
    pv.quiet_flag = 0;

	/* don't dump layer II header information by default */
	pv.show2hdr_flag = 0;

	/* don't dump application layer  data by default */
	pv.data_flag = 1;

	/* don't use UTC by default */
	pv.use_utc = 0;

#ifdef MPLS
    pv.mpls_payload_type = DEFAULT_MPLS_PAYLOADTYPE;
    pv.overlapping_IP = DEFAULT_MPLS_OVERLAPPING_IP;
    pv.mpls_stack_depth = DEFAULT_LABELCHAIN_LENGTH;
#endif
    
    /* chew up the command line */
    ParseCmdLine(argc, argv);

    /* If we are running non-root, install a dummy handler instead. */
    if (userid != 0)
        signal(SIGHUP, SigCantHupHandler);
    
    /* determine what run mode we are going to be in */
    if(pv.print_version)
    {
        /* Do nothing, just fall through */
        runMode = MODE_VERSION;
    }
	else if(pv.batch_mode_flag)
	{
		runMode = MODE_BATCH;
	}
    else
    {
		/* Check if a config file has already been inferred */
		if (!pv.config_file)
			pv.config_file = ConfigFileSearch();
		
        if (pv.config_file)
        {
            runMode = MODE_CONTINUOUS;
            if(!pv.quiet_flag)
            {
                LogMessage("Running in Continuous mode with inferred config file: %s\n", pv.config_file);
            }
        }
        else
        {
            /* unable to determine a run mode */
            ShowBanner();
            ShowUsage(progname);
            FatalError("\n\nUh, you need to tell me to do something...\n\n");
        }
    }
    
#ifdef DEBUG
	LogMessage("DEBUG is ENABLED\n");
#endif

    /* set the default logging dir if not set yet */
    /* XXX should probably be done after reading config files */
    if(!pv.log_dir)
    {
        pv.log_dir = SnortStrdup(DEFAULT_LOG_DIR);
    }
    
    /* extract the config directory from the config filename */
    if(pv.config_file)
    {
        /* is there a directory seperator in the filename */
        if(strrchr(pv.config_file,'/'))
        {
            char *tmp;
            /* lazy way, we waste a few bytes of memory here */
            pv.config_dir = SnortStrdup(pv.config_file);

            tmp = strrchr(pv.config_dir,'/');
            *(++tmp) = '\0';
        }
        else
        {
            pv.config_dir = SnortStrdup("./");
        }
        DEBUG_WRAP(DebugMessage(DEBUG_INIT, "Config file = %s, config dir = "
                    "%s\n", pv.config_file, pv.config_dir););
    }

	if (pv.use_utc == 1)
	{
		thiszone = 0;
	}
	else
	{
		/* set the timezone (ripped from tcpdump) */
		thiszone = gmt2local(0);
	}

    if(!pv.quiet_flag)
    {
        LogMessage("\n        --== Initializing Barnyard2 ==--\n");
    }
    /*
     * if daemon mode requested, fork daemon first, otherwise on linux
     * interface will be reset.
     */
    if(pv.daemon_flag)
    {
        DEBUG_WRAP(DebugMessage(DEBUG_INIT, "Entering daemon mode\n"););
        openlog("barnyard2", LOG_PID | LOG_CONS, LOG_DAEMON ); 
    }
    else if(pv.logtosyslog_flag)
    {
        openlog("barynard2", LOG_PID | LOG_CONS, LOG_DAEMON );
    }

    /*
     *  Display barnyard version information here so that we can also show dynamic
     *  plugin versions, if loaded.
     */
    switch ( pv.print_version )
    {
        case 'V':
            PrintVersion();
            exit(0);
            break;

		case '?':
            PrintVersion();
            ShowUsage(progname);
            if ( optopt )
                exit(1);
            exit(0);
            break;
    }

	/* initialise subsystems */
    InitProtoNames();
    InitInputPlugins();
	InitOutputPlugins();

	/* parse the config file */
	if (!(pv.quiet_flag && !pv.daemon_flag))
		LogMessage("Parsing rules files %s\n", pv.config_file);
	
	ParseRulesFile(pv.config_file, 0, PARSE_RULE_LINES);
  
	
	/* generate the maps */
	if (!(pv.quiet_flag && !pv.daemon_flag))
		LogMessage("Generating maps\n");

    /* ensure all map files have been specified */
    if ( pv.reference_file == NULL || pv.class_file == NULL ||
         pv.sid_msg_file == NULL || pv.gen_msg_file == NULL )
    {
            FatalError("Not all map files have been specified!\n"
                       "For meaningful operation classification, gen-msg, reference and sid-msg files\n"
                       "need to be defined either at the command line or in the configuration file.\n");
    }

    InitReferenceMap(pv.reference_file);
    InitClassMap(pv.class_file);
	InitSidMap(pv.sid_msg_file);
    InitGenMap(pv.gen_msg_file);

#ifdef DEBUG
	//
	// print after config file has been parsed since input plugins have some
	// extra configuration during this process. 
	//
	DumpInputPlugins();
	DumpOutputPlugins();
#endif

	/* check for waldo file usage */
	if (pv.waldo.state & WALDO_STATE_ENABLED)
	{
		int				ret;

		ret = spoolerReadWaldo(&pv.waldo);

		/* show waldo file contents on successful load */
		if (ret == WALDO_FILE_SUCCESS)
		{
			LogMessage("Using waldo file '%s':\n"
						"    spool directory = %s\n"
						"    spool filebase  = %s\n"
						"    time_stamp      = %lu\n"
						"    record_idx      = %lu\n", 
						pv.waldo.filepath,
						pv.waldo.data.spool_dir,
						pv.waldo.data.spool_filebase,
						pv.waldo.data.timestamp,
						pv.waldo.data.record_idx);
		}
		else if (ret == WALDO_FILE_EEXIST)
		{
			LogMessage("Using empty waldo file '%s'\n", pv.waldo.filepath);
		}
		else if (ret == WALDO_FILE_ETRUNC)
		{
			LogMessage("WARNING: Ignoring corrupt/truncated waldo"
						"file '%s'\n", pv.waldo.filepath);
		}
	}

	/* daemonise ourselves */
    if( pv.daemon_flag )
    {
        GoDaemon();
    }

    /* create the PID file */
    /* TODO should be part of the GoDaemon process */
    if((runMode == MODE_CONTINUOUS) || pv.daemon_flag || *pv.pidfile_suffix || pv.create_pid_file)
    {
        /* ... then create a PID file if not reading from a file */
        if (pv.daemon_flag || *pv.pidfile_suffix || pv.create_pid_file)
        {
            /* build a relatively unique named pid file */
            if (pv.interface)
            {
                CreatePidFile(pv.interface);
            }
            else
            {
			    LogMessage("WARNING: Non-unique PID file being created. Define an interface"
				    		"directive to improve uniqueness.\n");

                CreatePidFile("");
            }

            if (pv.daemon_flag)
            {
                SignalWaitingParent();
            }
        }
    }

	/* drop privileges if requested, when initialization is done */
    SetUidGid();
    
	/* ensure post configuration of any plugins */
	PostConfigInitPlugins();

    if(!pv.quiet_flag)
    {
        LogMessage("\n        --== Initialization Complete ==--\n");
    }

    /* Tell 'em who wrote it, what "it" is, and give props */
    if(!pv.quiet_flag)
    {
        PrintVersion();
    }

	if(pv.test_mode_flag)
	{
		LogMessage("Barnyard2 successfully loaded config and plugins.\n");
		CleanExit(0);
	}

    if(pv.daemon_flag)
    {
        LogMessage("Barnyard2 initialization completed successfully (pid=%u)\n", getpid());
    }
    
	/* Grab the start time to calculate processing time */
	bzero((char *) &tz, sizeof(tz));
	gettimeofday(&starttime, &tz);

    /* Batch processing mode */
    if(runMode == MODE_BATCH)
    {
        int idx;
		if( !(pv.batch_total_files > 0) )
        {
            FatalError("No files to batch process!\n");
        }

        for(idx = 0; idx < pv.batch_total_files; idx++)
        {
			if(!pv.quiet_flag)
                LogMessage("Processing: %s\n", pv.batch_filelist[idx]);
            
			ProcessBatch("", pv.batch_filelist[idx]);
        }
        CleanExit(0);
    }
	/* Continual processing mode */
	else if (runMode == MODE_CONTINUOUS)
	{
	    ProcessContinuous(pv.waldo.data.spool_dir, pv.waldo.data.spool_filebase,
						  pv.waldo.data.record_idx, pv.waldo.data.timestamp);
	}

    /* If we are exiting because of a HUP, re-exec ourselves */
    if (hup_check() || pv.restart_flag)
    {
        if (pv.daemon_flag)
        {
            char *arg;
            int argIndex;
            for  (argIndex = 0; argIndex < argc; argIndex++)
            {
                arg = progargs[argIndex];
                if (!strncmp(arg, "--restart", 9))
                {
                    break;
                }
                if (!strncmp(arg, "-D", 2))
                {
                    /* Replace -D with --restart */
                    /* a probable memory leak - but we're exec()ing anyway */
                    progargs[argIndex] = SnortStrdup("--restart");
                    break;
                }
            }
        }

#ifdef PARANOID
        execv(progname, progargs);
#else
        execvp(progname, progargs);
#endif

        /* only get here if we failed to restart */
        LogMessage("Restarting %s failed: %s\n", progname, strerror(errno));
        if (pv.daemon_flag || pv.logtosyslog_flag)
        {
            closelog();
        }
        exit(-1);
    }
    
    if (pv.daemon_flag || pv.logtosyslog_flag)
    {
        closelog();
    }

    return 0;
}

static void PrintVersion()
{
    ShowBanner();
}

/*
 * Function: ShowUsage(char *)
 *
 * Purpose:  Display the program options and exit
 *
 * Arguments: progname => name of the program (argv[0])
 *
 * Returns: 0 => success
 */
int ShowUsage(char *program_name)
{
    fprintf(stdout, "USAGE: %s [-options]\n", program_name);
	fprintf(stdout, "\n");
	fprintf(stdout, "General Options:\n");
	fprintf(stdout, "        -?         Show this information\n");
    fprintf(stdout, "        -c <file>  Use configuration file <file>\n");
    fprintf(stdout, "        -g <gname> Run barnyard gid as <gname> group (or gid) after initialization\n");
    fprintf(stdout, "        -m <umask> Set umask = <umask>\n");
    fprintf(stdout, "        -q         Quiet. Don't show banner and status report\n");
    fprintf(stdout, "        -u <uname> Run barnyard uid as <uname> user (or uid) after initialization\n");
    fprintf(stdout, "        -v         Be verbose\n");
    fprintf(stdout, "        -C <file>  Read the classification map from <file>\n");
    fprintf(stdout, "        -D         Run barnyard in background (daemon) mode\n");
    fprintf(stdout, "        -G <file>  Read the gen-msg map from <file>\n");
    fprintf(stdout, "        -L <dir>   Write output files in <dir>\n");
    fprintf(stdout, "        -R <file>  Read the reference map from <file>\n");
    fprintf(stdout, "        -S <file>  Read the sid-msg map from <file>\n");
    fprintf(stdout, "        -T         Test and report on the current configuration\n");
    fprintf(stdout, "        -U         Use UTC for timestamps\n");
    fprintf(stdout, "        -V         Show version and exit\n");
	fprintf(stdout, "\n");
    fprintf(stdout, "Continual Processing Options:\n");
    fprintf(stdout, "        -a <dir>   Archive processed files to <dir>\n");
    fprintf(stdout, "        -f <base>  Use <base> as the base filename pattern\n");
    fprintf(stdout, "        -d <dir>   Spool files from <dir>\n");
    fprintf(stdout, "        -n         Only process new events\n");
    fprintf(stdout, "        -w <file>  Enable bookmarking using <file>\n");
	fprintf(stdout, "\n");
    fprintf(stdout, "Batch Processing Mode Options:\n");
    fprintf(stdout, "        -o         Enable batch processing mode\n");
	fprintf(stdout, "\n");
    fprintf(stdout, "Longname options and their corresponding single char version\n");
    fprintf(stdout, "   --reference <file>              Same as -R\n");
    fprintf(stdout, "   --classification <file>         Same as -C\n");
    fprintf(stdout, "   --gen-msg <file>                Same as -G\n");
    fprintf(stdout, "   --sid-msg <file>                Same as -S\n");
    fprintf(stdout, "   --pid-path <path>               Specify the path for the barnyard PID file\n");
    fprintf(stdout, "   --create-pidfile                Create PID file, even when not in Daemon mode\n");
    fprintf(stdout, "   --nolock-pidfile                Do not try to lock barnyard PID file\n");
    fprintf(stdout, "   --help                          Same as -?\n");
    fprintf(stdout, "   --version                       Same as -V\n");
	fprintf(stdout, "\n");
	 
	return 0;
}

/*
 * Function: ParseCmdLine(int, char *)
 *
 * Purpose:  Parse command line args
 *
 * Arguments: argc => count of arguments passed to the routine
 *            argv => 2-D character array, contains list of command line args
 *
 * Returns: 0 => success, 1 => exit on error
 */

static char *valid_options = "?c:d:g:m:qu:vC:G:L:R:S:TUVa:f:nw:Do";

#define LONGOPT_ARG_NONE 0
#define LONGOPT_ARG_REQUIRED 1
#define LONGOPT_ARG_OPTIONAL 2
static struct option long_options[] = {
   {"classification", LONGOPT_ARG_REQUIRED, NULL, 'C'},
   {"reference", LONGOPT_ARG_REQUIRED, NULL, 'R'},
   {"gen-msg", LONGOPT_ARG_REQUIRED, NULL, 'G'},
   {"sid-msg", LONGOPT_ARG_REQUIRED, NULL, 'C'},
   {"version", LONGOPT_ARG_NONE, NULL, 'V'},
   {"help", LONGOPT_ARG_NONE, NULL, '?'},
   {"pid-path", LONGOPT_ARG_REQUIRED, NULL, PID_PATH},
   {"create-pidfile", LONGOPT_ARG_NONE, NULL, CREATE_PID_FILE},
   {"nolock-pidfile", LONGOPT_ARG_NONE, NULL, NOLOCK_PID_FILE},
   {0, 0, 0, 0}
};

int ParseCmdLine(int argc, char *argv[])
{
    int ch;                         /* storage var for getopt info */
    int umaskchange = 1;
    int defumask = 0;
    int option_index = -1;
    int isName = 0;
    int i;

    DEBUG_WRAP(DebugMessage(DEBUG_INIT, "Parsing command line...\n"););

    /* just to be sane.. */
    username = NULL;
    groupname = NULL;
    pv.pidfile_suffix[0] = 0;
    pv.logtosyslog_flag = 0;
    
    /*
    **  Set this so we know whether to return 1 on invalid input.
    **  Barnyard2, like Snort, uses '?' for help and getopt uses '?' for telling us there
    **  was an invalid option, so we can't use that to tell invalid input.
    **  Instead, we check optopt and it will tell us.
    */
    optopt = 0;

    /* loop through each command line var and process it */
    while((ch = getopt_long(argc, argv, valid_options, long_options, &option_index)) != -1)
    {
        DEBUG_WRAP(DebugMessage(DEBUG_INIT, "Processing cmd line switch: %c\n", ch););
        switch(ch)
        {
            case PID_PATH:
                SnortStrncpy(pv.pid_path, optarg, STD_BUF);
                pv.create_pid_file = 1;
                break;
            case CREATE_PID_FILE:
                pv.create_pid_file = 1;
                break;
            case NOLOCK_PID_FILE:
                pv.nolock_pid_file = 1;
                break;
            
            case 'c': /* use configuration file x */
                pv.config_file = SnortStrdup(optarg);
                break;

            case 'd': /* read from spool dir x */
				if ( SnortSnprintf(pv.waldo.data.spool_dir, STD_BUF, "%s", optarg) != SNORT_SNPRINTF_SUCCESS )
					FatalError("barnyard2: spool directory too long\n");
                break;

            case 'g': /* setgid handler */
                if(groupname != NULL)
                    free(groupname);
                groupname = (char *)SnortAlloc((strlen(optarg) + 1) * sizeof(char));

                bcopy(optarg, groupname, strlen(optarg));

                isName = 0;
                for (i=0;i<strlen(groupname);i++)
                {
                    if (isdigit(groupname[i]) == 0)
                    {
                        isName = 1;
                        break;
                    }
                }
                if (((groupid = atoi(groupname)) == 0) || isName)
                {
                    gr = getgrnam(groupname);
                    if(gr == NULL)
                        FatalError("Group \"%s\" unknown\n", groupname);

                    groupid = (unsigned long)gr->gr_gid;
                }
                break;
            
			case 'm': /* set the umask for the output files */
                {
                    char *p;
                    long val = 0;

                    umaskchange = 0;

                    val = strtol(optarg, &p, 8);
                    if (*p != '\0' || val < 0 || (val & ~FILEACCESSBITS))
                    {
                        FatalError("bad umask %s\n", optarg);
                    }
                    else
                    {
                        defumask = val;
                    }
                }

                break;
                
            case 'q':  /* no stdout output mode */
                pv.quiet_flag = 1;
                break;

            case 'u':  /* setuid */
                username = (char *)SnortAlloc((strlen(optarg) + 1) * sizeof(char));

                bcopy(optarg, username, strlen(optarg));

                isName = 0;
                for (i=0;i<strlen(username);i++)
                {
                    if (isdigit(username[i]) == 0)
                    {
                        isName = 1;
                        break;
                    }
                }

                if (((userid = atoi(username)) == 0) || isName)
                {
                    pw = getpwnam(username);
                    if(pw == NULL)
                        FatalError("User \"%s\" unknown\n", username);

                    userid = pw->pw_uid;
                }
                else
                {
                    pw = getpwuid(userid);
                    if(pw == NULL)
                        FatalError(
                                "Can not obtain username for uid: %lu\n",
                                (u_long) userid);
                }

                if(groupname == NULL)
                {
                    char name[256];

                    SnortSnprintf(name, 255, "%lu", (u_long) pw->pw_gid);

                    groupname = (char *)SnortAlloc((strlen(name) + 1) * sizeof(char));
                    groupid = pw->pw_gid;
                }
                DEBUG_WRAP(DebugMessage(DEBUG_INIT, "UserID: %lu GroupID: %lu\n",
                    (unsigned long) userid, (unsigned long) groupid););
                break;

            case 'v': /* be verbose */
                DEBUG_WRAP(DebugMessage(DEBUG_INIT, "Verbose Flag active\n"););
                pv.verbose_flag = 1;
                break;

            case 'C': /* read from classfication file x */
                pv.class_file = SnortStrdup(optarg);
                break;
            
            case 'R': /* read from reference file x */
                pv.reference_file = SnortStrdup(optarg);
                break;
            
			case 'G': /* read from gen-msg file x */
                pv.gen_msg_file = SnortStrdup(optarg);
                break;

            case 'L': /* use log dir <X> */
                pv.log_dir = SnortStrdup(optarg);

                if(access(pv.log_dir, 2) != 0)
                {
                    FatalError("log directory '%s' does not exist\n", 
                            pv.log_dir);
                }
                break;

            case 'M':
                pv.logtosyslog_flag = 1;
                break;
            
			case 'S': /* read from sid-msg file x */
                pv.sid_msg_file = SnortStrdup(optarg);
                break;

            case 'T': /* test mode, verify that the rules load properly */
                pv.test_mode_flag = 1;
                DEBUG_WRAP(DebugMessage(DEBUG_INIT, "Barnyard starting in test mode...\n"););
                if (pv.daemon_flag)
                {
                    FatalError("Cannot use test mode and daemon mode together."
                            "\nTo verify configuration run first in test "
                            "mode and then restart in daemon mode\n");
                }
                break;    

			case 'U': /* Use UTC for timestamps */
				pv.use_utc = 1;
				break;

            case 'V': /* prog ver already gets printed out, so we just exit */
            case '?':  /* show help and exit with 1 */
                pv.print_version = ch;
                pv.quiet_flag = 1;
				break;

			case 'a': /* set the archive after processing dir as x */
			    pv.archive_dir = SnortStrdup(optarg);
				break;

			case 'f': /* set the file base for reading */
				if ( SnortSnprintf(pv.waldo.data.spool_filebase, STD_BUF, "%s", optarg) != SNORT_SNPRINTF_SUCCESS )
					FatalError("barnyard2: spool filebase too long\n");
				break;

            case 'n': /* process only new records */
                DEBUG_WRAP(DebugMessage(DEBUG_INIT, "Process new records only flag active\n"););
                pv.new_records_only_flag = 1;
                break;

			case 'w': /* use waldo file x */
				if ( SnortSnprintf(pv.waldo.filepath, STD_BUF, "%s", optarg) != SNORT_SNPRINTF_SUCCESS )
					FatalError("barnyard2: waldo filepath too long\n");
				
				/* enable waldo file usage */
				pv.waldo.state |= WALDO_STATE_ENABLED;
				break;

            case 'D':                /* daemon mode */
                DEBUG_WRAP(DebugMessage(DEBUG_INIT, "Daemon mode flag active\n"););
                pv.daemon_flag = 1;
                pv.quiet_flag = 1;

                if (pv.test_mode_flag)
                {
                    FatalError("Cannot use test mode and daemon mode together."
                               "\nTo verify configuration run first in test "
                               "mode and then restart in daemon mode\n");
                }

                break;

            case 'o':
				DEBUG_WRAP(DebugMessage(DEBUG_INIT, "Batch processing flag active\n"););
                pv.batch_mode_flag = 1;
                break;
    
        }
    }

	/* when batch processing check for any remaining arguments which should */
	/* be a parsed as a list of files to process. */
	if (pv.batch_mode_flag && (optind < argc))
	{
		int idx = 0;

		pv.batch_total_files = argc - optind;
		pv.batch_filelist = SnortAlloc(pv.batch_total_files * sizeof(char *));

		while (optind < argc)
		{
			DEBUG_WRAP(DebugMessage(DEBUG_INIT, "Extra args: %s\n", argv[optind]););
			pv.batch_filelist[idx] = SnortStrdup(argv[optind]);

			idx++;
			optind++;
		}
		
		DEBUG_WRAP(DebugMessage(DEBUG_INIT, "Total files: %i\n", pv.batch_total_files););
	}

    /* if the umask arg happened, set umask */
    if (umaskchange)
    {
        umask(077);           /* set default to be sane */
    }
    else
    {
        umask(defumask);
    }

    return 0;
}


/* locate one of the possible default config files */
/* allocates memory to hold filename */
static char *ConfigFileSearch()
{
    struct stat			st;
    int					i;
    char				*conf_files[]={"/etc/barnyard2.conf", "./barnyard2.conf",
										"/etc/barnyard2/barnyard2.conf", NULL};
    char				*fname = NULL;
    char				*rval = NULL;

    i = 0;

    /* search the default set of config files */
    while(conf_files[i])
    {
        fname = conf_files[i];

        if(stat(fname, &st) != -1)
        {
            rval = SnortStrdup(fname);
            break;
        }
        i++;
    }

    /* search for .barnyard2rc in the HOMEDIR */
    if(!rval)
    {
        char *home_dir = NULL;

        if((home_dir = getenv("HOME")) != NULL)
        {
            char *snortrc = "/.barnyard2rc";
            int path_len;

            path_len = strlen(home_dir) + strlen(snortrc) + 1;

            /* create the full path */
            fname = (char *)SnortAlloc(path_len);

            SnortSnprintf(fname, path_len, "%s%s", home_dir, snortrc);

            if(stat(fname, &st) != -1)
                rval = fname;
            else
                free(fname);
        }
    }

    return rval;
}

/* Signal Handlers ************************************************************/
void saveSignalInfo(int signal)
{
    /* If we are already processing an exit signal, nothing to do */
    if (pv.exit_signal)
        return;

    pv.exit_signal = signal;
}

#ifdef CATCH_SEGV
static int sawSegv = 0;
static void SigSegvHandler(int signal)
{
    if (sawSegv == 1)
    {
        /* Handle the reentrant issues for SEGV...
         * If we get a SEGV while processing one, just exit */
        exit(-1);
    }

    sawSegv = 1;
    saveSignalInfo(signal);
    LogMessage("*** \n");
    LogMessage("*** Barnyard2 caught a SEGV exception, shutting down.\n");

    switch(signal_location)
    {
        case SIGLOC_PARSE_RULES_FILE:
            LogMessage("*** SEGV caught while parsing '%s' at line %d.\n",
                       file_name, file_line );
            break;

        case SIGLOC_PCAP_LOOP:
            if( current_otn )
            {
                LogMessage("*** SEGV caught in event generator id = %u, sid = %u\n",
                           current_otn->sigInfo.generator, current_otn->sigInfo.id);
            }
            break;

        default:
            break;
    }

    LogMessage("*** \n");

    CleanExit(1);
}
#endif

static void SigTermHandler(int signal)
{
    saveSignalInfo(signal);
}

static void SigIntHandler(int signal)
{
    saveSignalInfo(signal);
}   

static void SigQuitHandler(int signal)
{
    saveSignalInfo(signal);
}

static void SigHupHandler(int signal)
{
    pv.usr_signal = signal;
}

/**
 * dummy signal handler for nonroot users or chroot.
 *
 * @param signal signal to exec
 */
void SigCantHupHandler(int signal)
{
    LogMessage("Reload via Signal HUP does not work if you aren't root or are chroot'ed\n");
}

static void SigUsrHandler(int signal)
{
    /* Just set a flag that we caught a SIGUSR
     * (or whatever else we handle in here)
     */
    pv.usr_signal = signal;
}

void FreeProgVars()
{
    if (pv.config_dir)
        free(pv.config_dir);
	
	if (pv.hostname)
		free(pv.hostname);
	
	if (pv.interface)
		free(pv.interface);
	
	if (pv.pcap_cmd)
		free(pv.pcap_cmd);
	
    if (pv.config_file)
        free(pv.config_file);
    
//    if (pv.spool_dir)
//        free(pv.spool_dir);
    
//    if (pv.spool_filebase)
//        free(pv.spool_filebase);
    
    if(pv.log_dir)
        free(pv.log_dir);
    
    if (pv.gen_msg_file)
        free(pv.gen_msg_file);
    
    if (pv.sid_msg_file)
        free(pv.sid_msg_file);
    
    if (pv.class_file)
        free(pv.class_file);
    
    if (pv.reference_file)
        free(pv.reference_file);
    
    if (pv.archive_dir)
        free(pv.archive_dir);
    
	if (pv.batch_total_files > 0)
	{
		int idx;
		for(idx = 0; idx< pv.batch_total_files; idx++)
		{
			free(pv.batch_filelist[idx]);
		}
		free(pv.batch_filelist);
	}
}

/****************************************************************************
 *
 * Function: CleanExit()
 *
 * Purpose:  Clean up misc file handles and such and exit
 *
 * Arguments: exit value;
 *
 * Returns: void function
 *
 ****************************************************************************/
extern PluginSignalFuncNode *PluginShutdownList;
extern PluginSignalFuncNode *PluginCleanExitList;
extern PluginSignalFuncNode *PluginRestartList;

void CleanExit(int exit_val)
{
    PluginSignalFuncNode *idxPlugin = NULL;
    PluginSignalFuncNode *tempPlugin = NULL;

    /* This function can be called more than once.  For example,
     * once from the SIGINT signal handler, and once recursively
     * as a result of calling pcap_close() below.  We only need
     * to perform the cleanup once, however.  So the static
     * variable already_exiting will act as a flag to prevent
     * double-freeing any memory.  Not guaranteed to be
     * thread-safe, but it will prevent the simple cases.
     */
    static int already_exiting = 0;
    if( already_exiting != 0 )
    {
        return;
    }
    already_exiting = 1;

    /* Do some post processing on any incomplete Plugin Data */
    idxPlugin = PluginShutdownList;
    while(idxPlugin)
    {
        idxPlugin->func(SIGQUIT, idxPlugin->arg);
        tempPlugin = idxPlugin;
        idxPlugin = idxPlugin->next;
        free(tempPlugin);
    }

    if (!exit_val)
    {
        struct timeval difftime;
        struct timezone tz;

        bzero((char *) &tz, sizeof(tz));
        gettimeofday(&endtime, &tz);

        TIMERSUB(&endtime, &starttime, &difftime);

        if ( !pv.quiet_flag && !pv.test_mode_flag )
        {
            if (pv.done_processing)
            {
                printf("Run time for packet processing was %lu.%lu seconds\n", 
                    (unsigned long)difftime.tv_sec,
                    (unsigned long)difftime.tv_usec);
            }
            else if (pv.exit_signal)
            {
                printf("Run time prior to being shutdown was %lu.%lu seconds\n", 
                    (unsigned long)difftime.tv_sec,
                    (unsigned long)difftime.tv_usec);
            }
        }
    }

	/* Exit plugins */
	idxPlugin = PluginCleanExitList;
	while (idxPlugin)
	{
		idxPlugin->func(SIGQUIT, idxPlugin->arg);
		idxPlugin = idxPlugin->next;
	}

    /* free allocated memory */

    /* remove pid file */
    if(SnortStrnlen(pv.pid_filename, sizeof(pv.pid_filename)) > 0)
    {
        int ret;

        ret = unlink(pv.pid_filename);

        if (ret != 0)
        {
            ErrorMessage("Could not remove pid file %s: %s\n",
                         pv.pid_filename, strerror(errno));
        }
    }

    ClosePidFile();

    /* Free up stuff from pv... */
    FreeProgVars();

    LogMessage("Barnyard2 exiting\n");

    /* exit */
    exit(exit_val);
}

static void Restart(void)
{
    PluginSignalFuncNode *idxPlugin = NULL;
    PluginSignalFuncNode *plugTemp = NULL;

    /* Print statistics */
    if(!pv.test_mode_flag)
    {
//        fpShowEventStats();
//        DropStats(1);
    }

    /* Exit plugins */
    /* legacy exit code */
    idxPlugin = PluginRestartList;

    while(idxPlugin)
    {
        idxPlugin->func(SIGHUP, idxPlugin->arg);
        plugTemp = idxPlugin;
        idxPlugin = idxPlugin->next;
        free(plugTemp);
    }

    /* free allocated memory */
    

    /* remove pid file */
    if(SnortStrnlen(pv.pid_filename, sizeof(pv.pid_filename)) > 0)
    {
        int ret;

        ret = unlink(pv.pid_filename);

        if (ret != 0)
        {
            ErrorMessage("Could not remove pid file %s: %s\n",
                         pv.pid_filename, strerror(errno));
        }
    }

    /* Free up stuff from pv... */
    FreeProgVars();

    LogMessage("Restarting Barnyard2\n");

}

