/* 
**
** Copyright (C) 2008-2009 SecurixLive   <dev@securixlive.com>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License Version 2 as
** published by the Free Software Foundation.  You may not use, modify or
** distribute this program under any other version of the GNU General
** Public License.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
**
*/

/*
** Description:
**   In memory linked list structures of sid-msg.map, gen-msg.map and
** classification.config
**
** Author(s):
**   firnsy <firnsy@securixlive.com>
**   SecurixLive.com Team <dev@securixlive.com>
**
** Comments:
**   Ideas stolen liberally from:
**     1. the orginal barnyard (A. Baker, M. Roesch)
**
** TODO:
**   1. Convert existing linked lists to adaptive splayed trees.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#ifdef SOLARIS
    #include <strings.h>
#endif

#include "debug.h"
#include "map.h"
#include "util.h"
#include "mstring.h"
#include "strlcatu.h"

#include "barnyard2.h"

#include <string.h>
#include <stdlib.h>


/********************* Reference Implementation *******************************/

ReferenceNode *AddReference(ReferenceNode *rn, char *system, char *id)
{
    ReferenceNode *newNode;

    if(system == NULL || id == NULL)
    {
        ErrorMessage("NULL reference argument\n");
        return rn;
    }
    
    /* create the new node */
    newNode = (ReferenceNode *)SnortAlloc(sizeof(ReferenceNode));
    
    /* lookup the reference system */
    newNode->system = ReferenceSystemLookup(system);
    if (!newNode->system)
    {
        newNode->system = ReferenceSystemAdd(system, NULL);
    }
    newNode->id = SnortStrdup(id);
    
    DEBUG_WRAP(DebugMessage(DEBUG_MAPS, "map: adding reference (%s, %s)\n", system, id););
    
    /* add the node to the list */
    newNode->next = rn;
    
    return newNode;
}

/* print a reference node */
void FPrintReference(FILE *fp, ReferenceNode *refNode)
{
    if(refNode)
    {
        if(refNode->system)
        {
            if(refNode->system->url)
                fprintf(fp, "[Xref => %s%s]", refNode->system->url, 
                        refNode->id);
            else
                fprintf(fp, "[Xref => %s %s]", refNode->system->name,
                        refNode->id);
        }
        else
        {
            fprintf(fp, "[Xref => %s]", refNode->id);
        }
    }
    return;   
}

void ParseReference(char *args, SigNode *sn)
{
    char **toks, *system, *id;
    int num_toks;

    DEBUG_WRAP(DebugMessage(DEBUG_MAPS, "map: parsing reference %s\n", args););
    
    /* 2 tokens: system, id */
    toks = mSplit(args, ",", 2, &num_toks, 0);
    if(num_toks != 2)
    {
        LogMessage("WARNING: invalid Reference spec '%s'. Ignored\n", args);
    }
    else
    {
        system = toks[0];
        while ( isspace((int) *system) )
            system++;

        id = toks[1];
        while ( isspace((int) *id) )
            id++;
            
        sn->refs = AddReference(sn->refs, system, id);
    }

    mSplitFree(&toks, num_toks);

    return;
}


/********************* End of Reference Implementation ************************/

/********************** Reference System Implementation ***********************/

ReferenceSystemNode *referenceSystems = NULL;

ReferenceSystemNode *ReferenceSystemAdd(char *name, char *url)
{   
    ReferenceSystemNode *newNode;
    if(name == NULL)
    {
        ErrorMessage("NULL reference system name\n");
        return NULL;
    }

    /* create the new node */
    newNode = (ReferenceSystemNode *)SnortAlloc(sizeof(ReferenceSystemNode));

    newNode->name = SnortStrdup(name);
    if(url)
        newNode->url = SnortStrdup(url);
    else
        newNode->url = NULL;

    /* add to the list */
    newNode->next = referenceSystems;
    referenceSystems = newNode;
    return newNode;
}

void DeleteReferenceSystems()
{
    ReferenceSystemNode *current, *tmpReference;

    current = referenceSystems;
    while (current)
    {
        tmpReference = current->next;
        if (current->url)
            free(current->url);
        if (current->name)
            free(current->name);
        free(current);
        current = tmpReference;
    }
    referenceSystems = NULL;
}

ReferenceSystemNode *ReferenceSystemLookup(char *name)
{   
    ReferenceSystemNode *refSysNode = referenceSystems;
    while(refSysNode)
    {
        if(strcasecmp(name, refSysNode->name) == 0)
            return refSysNode;
        refSysNode = refSysNode->next;
    }
    return NULL;
}

void ParseReferenceSystemConfig(char *args)
{
    char **toks;
    char *name = NULL;
    char *url = NULL;
    int num_toks;

    /* 2 tokens: name <url> */
    toks = mSplit(args, " ", 2, &num_toks, 0);
    name = toks[0];
    if(num_toks == 2)
    {
        url = toks[1];
        while(isspace((int)*url))
            url++;
        if(url[0] == '\0')
            url = NULL;
    }
    ReferenceSystemAdd(name, url);

    mSplitFree(&toks, num_toks);
    return;
}

int InitReferenceMap(const char *file)
{
    FILE				*referencefile;
	int					ret;

    DEBUG_WRAP(DebugMessage(DEBUG_MAPS, "map: opening file %s\n", file););
    
    if((referencefile = fopen(file, "r")) == NULL)
    {
        LogMessage("ERROR: Unable to open Reference file '%s' (%s)\n", 
                file, strerror(errno));
        
        return -1;
    }

	ret = ReadReferenceMapFile(referencefile);

	if(referencefile != NULL)
		fclose(referencefile);

	return 0;
}

int ReadReferenceMapFile(FILE *fp)
{
    char				buf[BUFFER_SIZE];
    char				*index;
    char				**toks;
    int					num_toks;
	int					count = 0;

    bzero(buf, BUFFER_SIZE);
    
    while ( fgets(buf, BUFFER_SIZE, fp) != NULL )
    {
        index = buf;

        /* advance through any whitespace at the beginning of the line */
        while (*index == ' ' || *index == '\t')
            index++;

        /* if it's not a comment or a <CR>, send it to the parser */
        if ( (*index != '#') && (*index != 0x0a) && (index != NULL) )
        {
            toks = mSplit(index, ":", 2, &num_toks, 0);
            
            if(num_toks > 1)
            {
                ParseReferenceSystemConfig(toks[1]);
				count++;
            }

            mSplitFree(&toks, num_toks);
        }
    }

	return count;
}
/****************** End of Reference System Implementation ********************/


/************************ Class/Priority Implementation ***********************/

ClassType *classTypes = NULL;

int AddClassificationConfig(ClassType *newNode);

ClassType *GetClassByType(char *type)
{
    ClassType *idx = classTypes;
    if(!type)
        return NULL;

    while(idx)
    {
        if(strcasecmp(type, idx->type) == 0)
            return idx;
        idx = idx->next;
    }
    return NULL;
}

ClassType *GetClassById(u_int32_t id)
{
    ClassType *idx = classTypes;
    while(idx)
    {
        if(idx->id == id)
            return idx;
        idx = idx->next;
    }
    return NULL;
}


void ParseClassificationConfig(char *args)
{
    char **toks;
    int num_toks;
    char *data;
    ClassType *newNode;

    toks = mSplit(args, ",", 3, &num_toks, '\\');

    if(num_toks != 3)
    {
        ErrorMessage(": Invalid classification config: %s\n", args);
    }
    else
    {
        /* create the new node */
        newNode = (ClassType *)SnortAlloc(sizeof(ClassType));

        data = toks[0];
        while(isspace((int)*data)) 
            data++;
        newNode->type = SnortStrdup(data);   /* XXX: oom check */

        data = toks[1];
        while(isspace((int)*data))
            data++;
        newNode->name = SnortStrdup(data);   /* XXX: oom check */

        data = toks[2];
        while(isspace((int)*data))
            data++;
        /* XXX: error checking needed */
        newNode->priority = atoi(data); /* XXX: oom check */

        if(AddClassificationConfig(newNode) == -1)
        {
            ErrorMessage(": Duplicate classification \"%s\""
                    "found, ignoring this line\n", newNode->type);

            if(newNode)
            {
                if(newNode->name)
                    free(newNode->name);
                if(newNode->type)
                    free(newNode->type);
                free(newNode);
            }
        }
    }

    mSplitFree(&toks, num_toks);
    return;
}

void DeleteClassifications()
{
    ClassType *current, *tmpClass;

    current = classTypes;
    while (current)
    {
        tmpClass = current->next;
        if (current->type)
            free(current->type);
        if (current->name)
            free(current->name);
        free(current);
        current = tmpClass;
    }
    classTypes = NULL;
}

int AddClassificationConfig(ClassType *newNode)
{
    int max_id = 0;
    ClassType *current;

    current = classTypes;

    while(current)
    {
        /* dup check */
        if(strcasecmp(current->type, newNode->type) == 0)
            return -1;
        
        if(current->id > max_id)
            max_id = current->id;
        
        current = current->next;
    }

    /* insert node */
    newNode->id = max_id + 1;
    newNode->next = classTypes;
    classTypes = newNode;

    return newNode->id;
}

int InitClassMap(const char *file)
{
    FILE				*classfile;
	int					ret;

    DEBUG_WRAP(DebugMessage(DEBUG_MAPS, "map: opening file %s\n", file););
    
	if((classfile = fopen(file, "r")) == NULL)
    {
        LogMessage("ERROR: Unable to open Classification file '%s' (%s)\n", 
                file, strerror(errno));
        
        return -1;
    }

	ret = ReadClassMapFile(classfile);

	if(classfile != NULL)
		fclose(classfile);

	return 0;
}

int ReadClassMapFile(FILE *fp)
{
    char				buf[BUFFER_SIZE];
    char				*index;
    char				**toks;
    int					num_toks;
	int					count = 0;

    bzero(buf, BUFFER_SIZE);
    
    while ( fgets(buf, BUFFER_SIZE, fp) != NULL )
    {
        index = buf;

        /* advance through any whitespace at the beginning of the line */
        while (*index == ' ' || *index == '\t')
            index++;

        /* if it's not a comment or a <CR>, send it to the parser */
        if ( (*index != '#') && (*index != 0x0a) && (index != NULL) )
        {
            toks = mSplit(index, ":", 2, &num_toks, 0);
            
            if(num_toks > 1)
            {
                ParseClassificationConfig(toks[1]);
				count++;
            }

            mSplitFree(&toks, num_toks);
        }
    }

	return count;
}

/***************** End of Class/Priority Implementation ***********************/

/************************* Sid/Gid Map Implementation *************************/

SigNode					*sigTypes = NULL;

int InitSidMap(const char *file)
{
    FILE				*sid_map;
	int					ret;

    DEBUG_WRAP(DebugMessage(DEBUG_MAPS, "map: opening file %s\n", file););

    if( (sid_map=fopen(file, "r")) == NULL )
    {
        LogMessage("ERROR: Unable to open SID file '%s' (%s)\n", file, 
                strerror(errno));
        
        return -1;
    }

    ret = ReadSidMapFile(sid_map);

	if(sid_map != NULL)
		fclose(sid_map);

	return ret;
}


int ReadSidMapFile(FILE *fp)
{
    char				buf[BUFFER_SIZE];
    char				*index;
	int					count = 0;
    
    bzero(buf, BUFFER_SIZE);
    
    while(fgets(buf, BUFFER_SIZE, fp) != NULL)
    {
        index = buf;

        /* advance through any whitespace at the beginning of the line */
        while(*index == ' ' || *index == '\t')
            index++;

        /* if it's not a comment or a <CR>, send it to the parser */
        if((*index != '#') && (*index != 0x0a) && (index != NULL))
        {
            ParseSidMapLine(index);
			count++;
        }
    }

	return count;
}

void DeleteSigNodes()
{
    SigNode				*sn = NULL, *snn = NULL;
    ReferenceNode		*rn = NULL, *rnn = NULL;

	sn = sigTypes;

    while(sn)
    {
        snn = sn->next;
    
		/* free the message */
		if(sn->msg)
	        free(sn->msg);
    
		/* free the references (NOT the reference systems) */
	    if(sn->refs)
		{
			rn = sn->refs;
			while(rn)
		    {
		        rnn = rn->next;
		        
				/* free the id */
		        if(rn->id)
		            free(rn->id);
		        
				/* free the reference node */
				free(rn);

		        rn = rnn;
			}
		}

		/* free the signature node */
		free(sigTypes);

        sigTypes = snn;
    }

    sigTypes = NULL;
}

void ParseSidMapLine(char *data)
{
    char				**toks;
    int					num_toks;
    int					i;
    char				*idx;
    SigNode				*sn; 

    toks = mSplitSpecial(data, "||", 32, &num_toks, '\0');

    if(num_toks < 2)
    {
        LogMessage("WARNING: Ignoring bad line in SID file: '%s'\n", data);
    }
	else
	{
		DEBUG_WRAP(DebugMessage(DEBUG_MAPS, "map: creating new node\n"););

		sn = CreateSigNode(&sigTypes);
    
    	for(i = 0; i<num_toks; i++)
    	{
	        strip(toks[i]);
        	idx = toks[i];
        	while(*idx == ' ') idx++;
	            
        	switch(i)
        	{
	            case 0: /* sid */
                	sn->generator = 1;
                	sn->id = strtoul(idx, NULL, 10);
                	break;

	            case 1: /* msg */
                	sn->msg = SnortStrdup(idx);
                	break;

            	default: /* reference data */
					ParseReference(idx, sn);
					break;
			}
        }
    }

	mSplitFree(&toks, num_toks);
}

SigNode *GetSigByGidSid(u_int32_t gid, u_int32_t sid)
{
	/* set temp node pointer to the Sid map list head */
    SigNode *sn = sigTypes;
  
	/* a snort general rule (gid=1) and a snort dynamic rule (gid=3) use the  */
	/* the same sids and thus can be considered one in the same. */
	if (gid == 3)
		gid = 1;

    /* find any existing Snort ID's that match */
    while (sn != NULL)
    {
        if (sn->generator == gid && sn->id == sid)
        {
            return sn;
        }

        sn = sn->next;
    }

	/* create a default message since we didn't find any match */
	sn = CreateSigNode(&sigTypes);
    sn->generator = gid;
    sn->id = sid;
    sn->rev = 0;
    sn->msg = (char *)SnortAlloc(42);
    snprintf(sn->msg, 42, "Snort Alert [%u:%u:%u]", gid, sid, 0);
    
    return sn;
}

SigNode *CreateSigNode(SigNode **head)
{
    SigNode				*sn;

    if (*head == NULL)
    {
        *head = (SigNode *) SnortAlloc(sizeof(SigNode));
        return *head;
    }
    else
    {
        sn = *head;

        while (sn->next != NULL) 
			sn = sn->next;

        sn->next = (SigNode *) SnortAlloc(sizeof(SigNode));

        return sn->next;
    }
}

int InitGenMap(const char *file)
{
    FILE				*genfile;
	int					ret;

    if ( (genfile=fopen(file, "r")) == NULL )
    {
        LogMessage("ERROR: Unable to open Generator file \"%s\": %s\n", file, 
                strerror(errno));
        
        return -1;
    }

    ret = ReadGenMapFile(genfile);

	if(genfile != NULL)
		fclose(genfile);

	return 0;
}

int ReadGenMapFile(FILE *fp)
{
    char				buf[BUFFER_SIZE];
    char				*index;
	int					count = 0;
    
    bzero(buf, BUFFER_SIZE);
    
    while( fgets(buf, BUFFER_SIZE, fp) != NULL )
    {
        index = buf;

        /* advance through any whitespace at the beginning of the line */
        while (*index == ' ' || *index == '\t')
            index++;

        /* if it's not a comment or a <CR>, send it to the parser */
        if( (*index != '#') && (*index != 0x0a) && (index != NULL) )
        {
            ParseGenMapLine(index);
			count++;
        }
    }

	return count;
}

void ParseGenMapLine(char *data)
{
    char **toks;
    int num_toks;
    int i;
    char *idx;
    SigNode				*sn; 

    toks = mSplitSpecial(data, "||", 32, &num_toks, '\0');

    if(num_toks < 2)
    {
        LogMessage("WARNING: Ignoring bad line in SID file: \"%s\"\n", data);
		return;
    }

	sn = CreateSigNode(&sigTypes);
    
    for(i=0; i<num_toks; i++)
    {
        strip(toks[i]);
        idx = toks[i];
        while(*idx == ' ') idx++;
            
        switch(i)
        {
            case 0: /* gen */
				//TODO: error checking on conversion
                sn->generator = strtoul(idx, NULL, 10);
                break;

            case 1: /* sid */
				//TODO: error checking on conversion
                sn->id = strtoul(idx, NULL, 10);
                break;

            case 2: /* msg */
                sn->msg = SnortStrdup(idx);
                break;

            default: /* reference data */
				ParseReference(idx, sn);
                break;
        }
    }

    mSplitFree(&toks, num_toks);
}
