/* $Id$ */
/*
** Copyright (C) 2002-2009 Sourcefire, Inc.
** Copyright (C) 1998-2002 Martin Roesch <roesch@sourcefire.com>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License Version 2 as
** published by the Free Software Foundation.  You may not use, modify or
** distribute this program under any other version of the GNU General
** Public License.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/*
** Modifications by: Ian Firns <firnsy@securixlive.com>
**
** History:
**   2008-05-26 - Modifed GetUniqueName and GetHostname to provide better
**                handling of config file directives and lack thereof. 
**            
**                On 'nix hosts GetHostname will also attempt to determine the
**                current hostname of the box as well.
**
**   2008-05-09 - Cleaned and amalgamated the input and output plugin API into
**                the one file.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif

//#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <time.h>
#include <errno.h>

#include "plugbase.h"
#include "barnyard2.h"
#include "debug.h"
#include "util.h"

/* built-in input plugins */
#include "input-plugins/spi_unified2.h"

/* built-in output plugins */
#include "output-plugins/spo_alert_arubaaction.h"
#include "output-plugins/spo_alert_cef.h"
#include "output-plugins/spo_alert_fast.h"
#include "output-plugins/spo_alert_full.h"
#include "output-plugins/spo_alert_syslog.h"
#include "output-plugins/spo_alert_test.h"
#include "output-plugins/spo_alert_prelude.h"
#include "output-plugins/spo_alert_unixsock.h"
#include "output-plugins/spo_csv.h"
#include "output-plugins/spo_database.h"
#include "output-plugins/spo_log_ascii.h"
#include "output-plugins/spo_log_null.h"
#include "output-plugins/spo_log_tcpdump.h"
#include "output-plugins/spo_platypus.h"
#include "output-plugins/spo_sguil.h"


PluginSignalFuncNode *PluginShutdownList;
PluginSignalFuncNode *PluginCleanExitList;
PluginSignalFuncNode *PluginRestartList;
PluginSignalFuncNode *PluginPostConfigList;

extern int file_line;
extern char *file_name;

/***************************** Input Plugin API  *****************************/
InputKeywordList *InputKeywords;

InputFuncNode *InputList;

void InitInputPlugins()
{
    if(!pv.quiet_flag)
    {
        LogMessage("Initializing Input Plugins!\n");
    }
    
	Unified2Setup();
}

int ActivateInputPlugin(char *plugin_name, char *plugin_options)
{
    InputKeywordList *plugin;
    
    if(!plugin_name)
        return -1;
    
    /* get the intput plugin node */
    plugin = GetInputPlugin(plugin_name);
    if (!plugin)
        return -1;

    plugin->entry.func(plugin_options);

    return 0;
}

InputKeywordList *GetInputPlugin(char *plugin_name)
{
    InputKeywordList *list_node = NULL;
//    InputKeywordNode *ret = NULL;

    if (!plugin_name)
        return NULL;

    list_node = InputKeywords;

    while (list_node != NULL)
    {
        if (strcasecmp(plugin_name, list_node->entry.keyword) == 0)
        {
            //ret = &(list_node->entry);
            //ret = list_node;
            break;
        }

        list_node = list_node->next;
    }

    if (list_node == NULL)
    {
        FatalError("unknown input plugin: '%s'", plugin_name);
    }

    return list_node;
}

/****************************************************************************
 *
 * Function: RegisterInputPlugin(char *, void (*func)(Packet *, u_char *))
 *
 * Purpose:  Associates an input statement with its function. It will also
 * setup a InputFunctionNode of the same keyword which stores the 
 * apppropriate file processing methods.
 *
 * Arguments: keyword => The input keyword to associate with the
 *                       input processor
 *            *func => function pointer to the handler
 *
 * Returns: void function
 *
 ***************************************************************************/
void RegisterInputPlugin(char *keyword, InputInitFunc func)
{
    InputKeywordList	*idx;

    DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN,"Registering keyword:input.init() => %s:%p\n", 
                keyword, func););

	/* Establish the appropriate InputKeywordList reference */
    idx = InputKeywords;

    if(idx == NULL)
    {
        /* alloc the node */
        InputKeywords = (InputKeywordList *)SnortAlloc(sizeof(InputKeywordList));

        idx = InputKeywords;
    }
    else
    {
        /* loop to the end of the list to check for existence */
        while(idx->next != NULL)
        {
            if(!strcasecmp(idx->entry.keyword, keyword))
            {
                FatalError("%s(%d) => Duplicate input keyword!\n", 
                        file_name, file_line);
            }
            idx = idx->next;
        }

        idx->next = (InputKeywordList *)SnortAlloc(sizeof(InputKeywordList));

        idx = idx->next;
    }

    /* alloc space for the keyword */
    idx->entry.keyword = (char *)SnortAlloc((strlen(keyword) + 1) * sizeof(char));

    /* copy the keyword into the struct */
    SnortStrncpy(idx->entry.keyword, keyword, strlen(keyword) + 1);

    /* set the function pointer to the keyword handler function */
    idx->entry.func = func;

    /* the funcs entries are configured on plugin initilization */
}


/****************************************************************************
 *
 * Function: DumpInputPlugins()
 *
 * Purpose:  Prints the keyword->function list
 *
 * Arguments: None.
 *
 * Returns: void function
 *
 ***************************************************************************/
void DumpInputPlugins()
{
#ifdef DEBUG
    InputKeywordList *idx;

    if(pv.quiet_flag)
        return;

    idx = InputKeywords;

    DebugMessage(DEBUG_INIT,"-------------------------------------------------\n");
    DebugMessage(DEBUG_INIT," Keyword     | Input(s) @ \n");
    DebugMessage(DEBUG_INIT,"-------------------------------------------------\n");
    while(idx != NULL)
    {
        DebugMessage(DEBUG_INIT,"%-13s: init() = %p\n", 
						idx->entry.keyword, idx->entry.func);
		if (idx->configured_flag)
		{
			DebugMessage(DEBUG_INIT,"%-13s:   - readRecordHeader() = %p\n", 
							idx->entry.keyword, idx->funcs.readRecordHeader);
			DebugMessage(DEBUG_INIT,"%-13s:   - readRecord()       = %p\n", 
							idx->entry.keyword, idx->funcs.readRecord);
		}
        idx = idx->next;
    }
    DebugMessage(DEBUG_INIT,"-------------------------------------------------\n\n");
#endif
}

int AddArgToInputList(char *plugin_name, void *arg)
{
    InputKeywordList *plugin;
    
    if(!plugin_name)
        return -1;
    
    /* get the input plugin node */
    plugin = GetInputPlugin(plugin_name);
    if (!plugin)
        return -1;

    plugin->funcs.arg = arg;

	return 0;
}

int AddReadRecordHeaderFuncToInputList(char *plugin_name, int (*readRecordHeader)(void *))
{
    InputKeywordList *plugin;
    
    if(!plugin_name)
        return -1;
    
    /* get the input plugin node */
    plugin = GetInputPlugin(plugin_name);
    if (!plugin)
        return -1;

    plugin->funcs.readRecordHeader = readRecordHeader;
	plugin->configured_flag = InputFuncNodeConfigured(&(plugin->funcs));
	
	return 0;
}

int AddReadRecordFuncToInputList(char *plugin_name, int (*readRecord)(void *))
{
    InputKeywordList *plugin;
    
    if(!plugin_name)
        return -1;
    
    /* get the input plugin node */
    plugin = GetInputPlugin(plugin_name);
    if (!plugin)
        return -1;

    plugin->funcs.readRecord = readRecord;
	plugin->configured_flag = InputFuncNodeConfigured(&(plugin->funcs));
	
	return 0;
}

int InputFuncNodeConfigured(InputFuncNode *ifn)
{
	/* if not all functions are defined then return a zero flag */
	if (!ifn->readRecordHeader || !ifn->readRecord)
		return 0;

	return 1;
}

/*************************** End Input Plugin API  ***************************/

/***************************** Output Plugin API  *****************************/
OutputKeywordList *OutputKeywords;

OutputFuncNode *OutputList;

OutputFuncNode *AlertList;
OutputFuncNode *LogList;

OutputFuncNode *AppendOutputFuncList(void (*) (Packet *, void *, uint32_t, void *),
                void *, OutputFuncNode *);

void InitOutputPlugins()
{
    if(!pv.quiet_flag)
    {
        LogMessage("Initializing Output Plugins!\n");
    }
    AlertCEFSetup();
    AlertSyslogSetup();
    LogTcpdumpSetup();
    DatabaseSetup();
    AlertFastSetup();
    AlertFullSetup();
#ifndef WIN32
    /* Win32 doesn't support AF_UNIX sockets */
    AlertUnixSockSetup();
#endif /* !WIN32 */
    AlertCSVSetup();
    LogNullSetup();
    LogAsciiSetup();

#ifdef ARUBA
    AlertArubaActionSetup();
#endif

#ifdef HAVE_LIBPRELUDE
    AlertPreludeSetup();
#endif

    AlertTestSetup();

	PlatypusSetup();
	SguilSetup();
}

int ActivateOutputPlugin(char *plugin_name, char *plugin_options)
{
    OutputKeywordNode *plugin;
    
    if(!plugin_name)
        return -1;
    
    /* get the output plugin node */
    plugin = GetOutputPlugin(plugin_name);
    if (!plugin)
        return -1;

	switch(plugin->node_type)
	{
		case NT_OUTPUT_SPECIAL:
			plugin->func(plugin_options);
			break;
		case NT_OUTPUT_ALERT:
			plugin->func(plugin_options);
			break;
		case NT_OUTPUT_LOG:
			plugin->func(plugin_options);
			break;
	}

    return 0;
}

OutputKeywordNode *GetOutputPlugin(char *plugin_name)
{
    OutputKeywordList *list_node = NULL;
    OutputKeywordNode *ret = NULL;

    if (!plugin_name)
        return NULL;

    list_node = OutputKeywords;

    while (list_node != NULL)
    {
        if (strcasecmp(plugin_name, list_node->entry.keyword) == 0)
        {
            ret = &(list_node->entry);
            break;
        }

        list_node = list_node->next;
    }

    if (ret == NULL)
    {
        FatalError("unknown output plugin: '%s'", plugin_name);
    }

    return ret;
}


/****************************************************************************
 *
 * Function: RegisterOutputPlugin(char *, void (*func)(Packet *, u_char *))
 *
 * Purpose:  Associates an output statement with its function.
 *
 * Arguments: keyword => The output keyword to associate with the
 *                       output processor
 *            type => alert or log types
 *            *func => function pointer to the handler
 *
 * Returns: void function
 *
 ***************************************************************************/
void RegisterOutputPlugin(char *keyword, int type, OutputInitFunc func)
{
    OutputKeywordList *idx;

    DEBUG_WRAP(DebugMessage(DEBUG_PLUGIN,"Registering keyword:output => %s:%p\n", 
                keyword, func););

    idx = OutputKeywords;

    if(idx == NULL)
    {
        /* alloc the node */
        OutputKeywords = (OutputKeywordList *)SnortAlloc(sizeof(OutputKeywordList));

        idx = OutputKeywords;
    }
    else
    {
        /* loop to the end of the list */
        while(idx->next != NULL)
        {
            if(!strcasecmp(idx->entry.keyword, keyword))
            {
                FatalError("%s(%d) => Duplicate output keyword!\n", 
                        file_name, file_line);
            }
            idx = idx->next;
        }

        idx->next = (OutputKeywordList *)SnortAlloc(sizeof(OutputKeywordList));

        idx = idx->next;
    }

    /* alloc space for the keyword */
    idx->entry.keyword = (char *)SnortAlloc((strlen(keyword) + 1) * sizeof(char));

    /* copy the keyword into the struct */
    SnortStrncpy(idx->entry.keyword, keyword, strlen(keyword) + 1);

    /*
     * set the plugin type, needed to determine whether an overriding command
     * line arg has been specified
     */
    idx->entry.node_type = (char) type;

    /* set the function pointer to the keyword handler function */
    idx->entry.func = func;
}


/****************************************************************************
 *
 * Function: DumpOutputPlugins()
 *
 * Purpose:  Prints the Output keyword list
 *
 * Arguments: None.
 *
 * Returns: void function
 *
 ***************************************************************************/
void DumpOutputPlugins(void)
{
    OutputKeywordList *idx;

    if(pv.quiet_flag)
        return;

    idx = OutputKeywords;

    LogMessage("-------------------------------------------------\n");
    LogMessage(" Keyword     |          Output @ \n");
    LogMessage("-------------------------------------------------\n");
    while(idx != NULL)
    {
        LogMessage("%-13s:       %p\n", idx->entry.keyword, idx->entry.func);
        idx = idx->next;
    }
    LogMessage("-------------------------------------------------\n\n");
}

void AddFuncToOutputList(void (*func) (Packet *, void *, uint32_t, void *),
        char node_type, void *arg)
{

	switch(node_type)
	{
		case NT_OUTPUT_ALERT:
			AlertList = AppendOutputFuncList(func, arg, AlertList);
			break;
		case NT_OUTPUT_LOG:
			LogList = AppendOutputFuncList(func, arg, LogList);
			break;
		default:
			FatalError("Unknown nodetype: %i. Possible bug, please report.\n", node_type);
	}

	return;
}

OutputFuncNode *AppendOutputFuncList(
        void (*func) (Packet *, void *, uint32_t, void *),
        void *arg, OutputFuncNode *list)
{
    OutputFuncNode *idx = list;

    if(idx == NULL)
    {
        idx = (OutputFuncNode *)SnortAlloc(sizeof(OutputFuncNode));
        idx->func = func;
        idx->arg = arg;
        list = idx;
    }
    else
    {
        while(idx->next != NULL)
            idx = idx->next;

        idx->next = (OutputFuncNode *)SnortAlloc(sizeof(OutputFuncNode));
        idx = idx->next;
        idx->func = func;
        idx->arg = arg;
    }

    idx->next = NULL;

    return list;
}

void CallOutputPlugins(char node_type, Packet *packet, void *event, uint32_t event_type)
{
	OutputFuncNode *idx = NULL;

    if (node_type == NT_OUTPUT_SPECIAL)
    {
        idx = AlertList;
    	while (idx != NULL)
    	{
    		idx->func(packet, event, event_type, idx->arg);
    		idx = idx->next;
    	}

        idx = LogList;
    	while (idx != NULL)
    	{
    		idx->func(packet, event, event_type, idx->arg);
    		idx = idx->next;
    	}
    }
    else
    {
    	switch(node_type)
    	{
    		case NT_OUTPUT_ALERT:
    			idx = AlertList;
    			break;
    		case NT_OUTPUT_LOG:
    			idx = LogList;
    			break;
    	}

    	while (idx != NULL)
    	{
    		idx->func(packet, event, event_type, idx->arg);
    		idx = idx->next;
    	}
    }
}

/*************************** End Output Plugin API  ***************************/


/************************** Miscellaneous Functions  **************************/

void PostConfigInitPlugins()
{
	PluginSignalFuncNode *idx;

	idx = PluginPostConfigList;

	while (idx != NULL)
	{
		idx->func(0, idx->arg);
		idx = idx->next;
	}
}

/* functions to aid in cleaning up after plugins */
void AddFuncToRestartList(void (*func) (int, void *), void *arg)
{
    PluginRestartList = AddFuncToSignalList(func, arg, PluginRestartList);
}

void AddFuncToCleanExitList(void (*func) (int, void *), void *arg)
{
    PluginCleanExitList = AddFuncToSignalList(func, arg, PluginCleanExitList);
}

void AddFuncToShutdownList(void (*func) (int, void *), void *arg)
{
    PluginShutdownList = AddFuncToSignalList(func, arg, PluginShutdownList);
}

void AddFuncToPostConfigList(void (*func)(int, void *), void *arg)
{
    PluginPostConfigList = AddFuncToSignalList(func, arg, PluginPostConfigList);
}

PluginSignalFuncNode *AddFuncToSignalList(void (*func) (int, void *), void *arg,
                                          PluginSignalFuncNode * list)
{
    PluginSignalFuncNode *idx;

    idx = list;

    if(idx == NULL)
    {
        idx = (PluginSignalFuncNode *)SnortAlloc(sizeof(PluginSignalFuncNode));

        idx->func = func;
        idx->arg = arg;
        list = idx;
    }
    else
    {
        while(idx->next != NULL)
            idx = idx->next;

        idx->next = (PluginSignalFuncNode *)SnortAlloc(sizeof(PluginSignalFuncNode));

        idx = idx->next;
        idx->func = func;
        idx->arg = arg;
    }
    idx->next = NULL;

    return list;
}


/****************************************************************************
 *
 * Function: GetUniqueName(char * iface)
 *
 * Purpose: To return a string that has a high probability of being unique
 *          for a given sensor.
 *
 * Arguments: char * iface - The network interface you are sniffing
 *
 * Returns: A char * -- its a static char * so you should not free it
 *
 ***************************************************************************/
char *GetUniqueName()
{
    char * rptr;
    static char uniq_name[256];

	if (pv.interface)
	{
		SnortSnprintf(uniq_name, 255, "%s:%s\n",GetHostname(), pv.interface);
	}
	else 
	{
	    SnortSnprintf(uniq_name, 255, "%s\n",GetHostname());
		LogMessage("No interface directive is defined. "
					"Name may not be unique for the host.\n");
    }

    rptr = uniq_name; 
    
    if (pv.verbose_flag) LogMessage("Node unique name is: %s\n", rptr);
    return rptr;
}    

/****************************************************************************
 *
 * Function: GetHostname()
 *
 * Purpose: To return a string representing the hostname
 *
 * Arguments: None
 *
 * Returns: A static char * representing the hostname. 
 *
 ***************************************************************************/
char *GetHostname()
{
    char				*error = "unknown";
	static char			hostname[255];

	if (pv.hostname) 
		return pv.hostname;
	else if (gethostname(hostname, 245) == 0)
		return hostname;
    else if(getenv("HOSTNAME")) 
		return getenv("HOSTNAME");
    else if(getenv("HOST")) 
		return getenv("HOST");
    else 
		return error;
}

/****************************************************************************
 *
 * Function: GetTimestamp(register const struct timeval *tvp, int tz)
 *
 * Purpose: Get an ISO-8601 formatted timestamp for tvp within the tz
 *          timezone. 
 *
 * Arguments: sec is time in seconds.
 *            usec is microsecond component.
 *            tz is a timezone. 
 *
 * Returns: char * -- You must free this char * when you are done with it.
 *
 ***************************************************************************/
char *GetTimestampByComponent(u_int32_t sec, u_int32_t usec, int tz)
{
    struct tm			*lt;  /* localtime */
    char				*buf;
	time_t				Time;
    int					msec;

	buf = (char *)SnortAlloc(SMALLBUFFER * sizeof(char));
    
	Time = sec;
    msec = usec / 1000;

	if (pv.use_utc == 1)
	{
		lt = gmtime(&Time);
        SnortSnprintf(buf, SMALLBUFFER,
                "%04i-%02i-%02i %02i:%02i:%02i.%03i",
                1900 + lt->tm_year, lt->tm_mon + 1, lt->tm_mday,
                lt->tm_hour, lt->tm_min, lt->tm_sec, msec);
	}
	else
	{
		lt = localtime(&Time);
        SnortSnprintf(buf, SMALLBUFFER,
                "%04i-%02i-%02i %02i:%02i:%02i.%03i+%03i",
                1900 + lt->tm_year, lt->tm_mon + 1, lt->tm_mday,
                lt->tm_hour, lt->tm_min, lt->tm_sec, msec, tz);
	}

    return buf;
}

/****************************************************************************
 *
 * Function: GetTimestampByStruct(register const struct timeval *tvp, int tz)
 *
 * Purpose: Get an ISO-8601 formatted timestamp for tvp within the tz
 *          timezone. 
 *
 * Arguments: tvp is a timeval pointer. tz is a timezone. 
 *
 * Returns: char * -- You must free this char * when you are done with it.
 *
 ***************************************************************************/
char *GetTimestampByStruct(register const struct timeval *tvp, int tz)
{
    struct tm *lt;  /* localtime */
    char * buf;
    int msec;

    buf = (char *)SnortAlloc(SMALLBUFFER * sizeof(char));

    msec = tvp->tv_usec / 1000;

    if(pv.use_utc == 1)
    {
        lt = gmtime((time_t *)&tvp->tv_sec);
        SnortSnprintf(buf, SMALLBUFFER, "%04i-%02i-%02i %02i:%02i:%02i.%03i",
                1900 + lt->tm_year, lt->tm_mon + 1, lt->tm_mday,
                lt->tm_hour, lt->tm_min, lt->tm_sec, msec);
    }
    else
    {
        lt = localtime((time_t *)&tvp->tv_sec);
        SnortSnprintf(buf, SMALLBUFFER,
                "%04i-%02i-%02i %02i:%02i:%02i.%03i+%03i",
                1900 + lt->tm_year, lt->tm_mon + 1, lt->tm_mday,
                lt->tm_hour, lt->tm_min, lt->tm_sec, msec, tz);
    }

    return buf;
}

/****************************************************************************
 *
 * Function: GetLocalTimezone()
 *
 * Purpose: Find the offset from GMT for current host
 *
 * Arguments: none 
 *
 * Returns: int representing the offset from GMT
 *
 ***************************************************************************/
int GetLocalTimezone()
{
    time_t      ut;
    struct tm * ltm;
    long        seconds_away_from_utc;

    time(&ut);
    ltm = localtime(&ut);

#if defined(SOLARIS) || defined(AIX) || defined(HPUX)
    /* localtime() sets the global timezone variable,
       which is defined in <time.h> */
    seconds_away_from_utc = timezone;
#else
    seconds_away_from_utc = ltm->tm_gmtoff;
#endif

    return  seconds_away_from_utc/3600;
}

/****************************************************************************
 *
 * Function: GetCurrentTimestamp()
 *
 * Purpose: Generate an ISO-8601 formatted timestamp for the current time.
 *
 * Arguments: none 
 *
 * Returns: char * -- You must free this char * when you are done with it.
 *
 ***************************************************************************/
char *GetCurrentTimestamp()
{
    struct tm *lt;
    struct timezone tz;
    struct timeval tv;
    struct timeval *tvp;
    char * buf;
    int tzone;
    int msec;

    buf = (char *)SnortAlloc(SMALLBUFFER * sizeof(char));

    bzero((char *)&tz,sizeof(tz));
    gettimeofday(&tv,&tz);
    tvp = &tv;

    msec = tvp->tv_usec/1000;

    if(pv.use_utc == 1)
    {
        lt = gmtime((time_t *)&tvp->tv_sec);
        SnortSnprintf(buf, SMALLBUFFER, "%04i-%02i-%02i %02i:%02i:%02i.%03i",
                1900 + lt->tm_year, lt->tm_mon + 1, lt->tm_mday,
                lt->tm_hour, lt->tm_min, lt->tm_sec, msec);
    }
    else
    {
        lt = localtime((time_t *)&tvp->tv_sec);

        tzone = GetLocalTimezone();

        SnortSnprintf(buf, SMALLBUFFER,
                "%04i-%02i-%02i %02i:%02i:%02i.%03i+%03i",
                1900 + lt->tm_year, lt->tm_mon + 1, lt->tm_mday,
                lt->tm_hour, lt->tm_min, lt->tm_sec, msec, tzone);
    }

    return buf;
}

/****************************************************************************
 * Function: base64(char * xdata, int length)
 *
 * Purpose: Insert data into the database
 *
 * Arguments: xdata  => pointer to data to base64 encode
 *            length => how much data to encode 
 *
 * Make sure you allocate memory for the output before you pass
 * the output pointer into this function. You should allocate 
 * (1.5 * length) bytes to be safe.
 *
 * Returns: data base64 encoded as a char *
 *
 ***************************************************************************/
char * base64(const u_char * xdata, int length)
{
    int count, cols, bits, c, char_count;
    unsigned char alpha[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";  /* 64 bytes */
    char * payloadptr;
    char * output;
    char_count = 0;
    bits = 0;
    cols = 0;

    output = (char *)SnortAlloc( ((unsigned int) (length * 1.5 + 4)) * sizeof(char) );

    payloadptr = output;

    for(count = 0; count < length; count++)
    {
        c = xdata[count];

        if(c > 255)
        {
            ErrorMessage("plugbase.c->base64(): encountered char > 255 (decimal %d)\n If you see this error message a char is more than one byte on your machine\n This means your base64 results can not be trusted", c);
        }

        bits += c;
        char_count++;

        if(char_count == 3)
        {
            *output = alpha[bits >> 18]; output++;
            *output = alpha[(bits >> 12) & 0x3f]; output++;
            *output = alpha[(bits >> 6) & 0x3f]; output++;
            *output = alpha[bits & 0x3f]; output++; 
            cols += 4;
            if(cols == 72)
            {
                *output = '\n'; output++;
                cols = 0;
            }
            bits = 0;
            char_count = 0;
        }
        else
        {
            bits <<= 8;
        }
    }

    if(char_count != 0)
    {
        bits <<= 16 - (8 * char_count);
        *output = alpha[bits >> 18]; output++;
        *output = alpha[(bits >> 12) & 0x3f]; output++;
        if(char_count == 1)
        {
            *output = '='; output++;
            *output = '='; output++;
        }
        else
        {
            *output = alpha[(bits >> 6) & 0x3f]; 
            output++; *output = '='; 
            output++;
        }
    }
    *output = '\0';
    return payloadptr;
} 

/****************************************************************************
 *
 * Function: ascii(u_char *xdata, int length)
 *
 * Purpose: This function takes takes a buffer "xdata" and its length then
 *          returns a string of only the printable ASCII characters.
 *
 * Arguments: xdata is the buffer, length is the length of the buffer in
 *            bytes
 *
 * Returns: char * -- You must free this char * when you are done with it.
 *
 ***************************************************************************/
char *ascii(const u_char *xdata, int length)
{
     char *d_ptr, *ret_val;
     int i,count = 0;
     int size;
     
     if(xdata == NULL)
     {
         return NULL;         
     }
     
     for(i=0;i<length;i++)
     {
         if(xdata[i] == '<')
             count+=4;              /* &lt; */
         else if(xdata[i] == '&')
             count+=5;              /* &amp; */
         else if(xdata[i] == '>')   /* &gt;  */
             count += 4;
     }

     size = length + count + 1;
     ret_val = (char *) calloc(1,size);
     
     if(ret_val == NULL)
     {
         LogMessage("plugbase.c: ascii(): Out of memory, can't log anything!\n");
         return NULL;
     }
     
     d_ptr = ret_val; 
     
     for(i=0;i<length;i++)
     {
         if((xdata[i] > 0x1F) && (xdata[i] < 0x7F))
         {
             if(xdata[i] == '<')
             {
                 SnortStrncpy(d_ptr, "&lt;", size - (d_ptr - ret_val));
                 d_ptr+=4;
             }
             else if(xdata[i] == '&')
             {
                 SnortStrncpy(d_ptr, "&amp;", size - (d_ptr - ret_val));
                 d_ptr += 5;
             }
             else if(xdata[i] == '>')
             {
                 SnortStrncpy(d_ptr, "&gt;", size - (d_ptr - ret_val));
                 d_ptr += 4;
             }
             else
             {
                 *d_ptr++ = xdata[i];
             }
         }
         else
         {
             *d_ptr++ = '.';
         }        
     }
     
     *d_ptr++ = '\0';
     
     return ret_val;
}

/****************************************************************************
 *
 * Function: hex(u_char *xdata, int length)
 *
 * Purpose: This function takes takes a buffer "xdata" and its length then
 *          returns a string of hex with no spaces
 *
 * Arguments: xdata is the buffer, length is the length of the buffer in
 *            bytes
 *
 * Returns: char * -- You must free this char * when you are done with it.
 *
 ***************************************************************************/
char *hex(const u_char *xdata, int length)
{
    int x;
    char *rval = NULL;
    char *buf = NULL;

    if (xdata == NULL)
        return NULL;

    buf = (char *)calloc((length * 2) + 1, sizeof(char));

    if (buf != NULL)
    {
        rval = buf;

        for (x = 0; x < length; x++)
        {
            SnortSnprintf(buf, 3, "%02X", xdata[x]);
            buf += 2;
        } 

        rval[length * 2] = '\0';
    }

    return rval;
}



char *fasthex(const u_char *xdata, int length)
{
    char conv[] = "0123456789ABCDEF";
    char *retbuf = NULL; 
    const u_char *index;
    const u_char *end;
    char *ridx;

    index = xdata;
    end = xdata + length;
    retbuf = (char *)SnortAlloc(((length * 2) + 1) * sizeof(char));
    ridx = retbuf;

    while(index < end)
    {
        *ridx++ = conv[((*index & 0xFF)>>4)];
        *ridx++ = conv[((*index & 0xFF)&0x0F)];
        index++;
    }

    return retbuf;
}

